@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import abc
 import asyncio
 import concurrent.futures as concurrent
@@ -11,15 +13,51 @@ import time
 import traceback
 import warnings
 from contextlib import suppress
+from datetime import datetime, timedelta
+from typing import TYPE_CHECKING, Any, Callable, Union
 
 import loky
 
+from adaptive import (
+    BalancingLearner,
+    BaseLearner,
+    DataSaver,
+    IntegratorLearner,
+    SequenceLearner,
+)
 from adaptive.notebook_integration import in_ipynb, live_info, live_plot
+from adaptive.utils import SequentialExecutor
+
+ExecutorTypes: TypeAlias = Union[
+    concurrent.ProcessPoolExecutor,
+    concurrent.ThreadPoolExecutor,
+    SequentialExecutor,
+    loky.reusable_executor._ReusablePoolExecutor,
+]
+FutureTypes: TypeAlias = Union[concurrent.Future, asyncio.Future, asyncio.Task]
+
+if TYPE_CHECKING:
+    import holoviews
+
+try:
+    from typing import TypeAlias
+except ImportError:
+    from typing_extensions import TypeAlias
+
+try:
+    from typing import Literal
+except ImportError:
+    from typing_extensions import Literal
 
 try:
     import ipyparallel
+    from ipyparallel.client.asyncresult import AsyncResult
 
     with_ipyparallel = True
+    ExecutorTypes: TypeAlias = Union[
+        ExecutorTypes, ipyparallel.Client, ipyparallel.client.view.ViewExecutor
+    ]
+    FutureTypes: TypeAlias = Union[FutureTypes, AsyncResult]
 except ModuleNotFoundError:
     with_ipyparallel = False
 
@@ -27,6 +65,9 @@ try:
     import distributed
 
     with_distributed = True
+    ExecutorTypes: TypeAlias = Union[
+        ExecutorTypes, distributed.Client, distributed.cfexecutor.ClientExecutor
+    ]
 except ModuleNotFoundError:
     with_distributed = False
 
@@ -34,16 +75,18 @@ try:
     import mpi4py.futures
 
     with_mpi4py = True
+    ExecutorTypes: TypeAlias = Union[ExecutorTypes, mpi4py.futures.MPIPoolExecutor]
 except ModuleNotFoundError:
     with_mpi4py = False
 
-
 with suppress(ModuleNotFoundError):
     import uvloop
 
     asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
 
 
+# -- Runner definitions
+
 if platform.system() == "Linux":
     _default_executor = concurrent.ProcessPoolExecutor
 else:
@@ -57,15 +100,31 @@ else:
 
 
 class BaseRunner(metaclass=abc.ABCMeta):
-    r"""Base class for runners that use `concurrent.futures.Executors`.
+    r"""Base class for runners that use `concurrent.futures.Executor`\'s.
 
     Parameters
     ----------
     learner : `~adaptive.BaseLearner` instance
-    goal : callable
+    goal : callable, optional
         The end condition for the calculation. This function must take
         the learner as its sole argument, and return True when we should
         stop requesting more points.
+    loss_goal : float, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the loss is smaller than this value.
+    npoints_goal : int, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the number of points is larger or
+        equal than this value.
+    end_time_goal : datetime, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than this
+        value.
+    duration_goal : timedelta or number, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than
+        ``start_time + duration_goal``. ``duration_goal`` can be a number
+        indicating the number of seconds.
     executor : `concurrent.futures.Executor`, `distributed.Client`,\
                `mpi4py.futures.MPIPoolExecutor`, `ipyparallel.Client` or\
                `loky.get_reusable_executor`, optional
@@ -87,6 +146,8 @@ class BaseRunner(metaclass=abc.ABCMeta):
         the point is present in ``runner.failed``.
     raise_if_retries_exceeded : bool, default: True
         Raise the error after a point ``x`` failed `retries`.
+    allow_running_forever : bool, default: False
+        Allow the runner to run forever when the goal is None.
 
     Attributes
     ----------
@@ -114,53 +175,66 @@ class BaseRunner(metaclass=abc.ABCMeta):
 
     def __init__(
         self,
-        learner,
-        goal,
+        learner: BaseLearner,
+        goal: Callable[[BaseLearner], bool] | None = None,
         *,
-        executor=None,
-        ntasks=None,
-        log=False,
-        shutdown_executor=False,
-        retries=0,
-        raise_if_retries_exceeded=True,
+        loss_goal: float | None = None,
+        npoints_goal: int | None = None,
+        end_time_goal: datetime | None = None,
+        duration_goal: timedelta | int | float | None = None,
+        executor: (ExecutorTypes | None) = None,
+        ntasks: int = None,
+        log: bool = False,
+        shutdown_executor: bool = False,
+        retries: int = 0,
+        raise_if_retries_exceeded: bool = True,
+        allow_running_forever: bool = False,
     ):
 
         self.executor = _ensure_executor(executor)
-        self.goal = goal
+        self.goal = _goal(
+            learner,
+            goal,
+            loss_goal,
+            npoints_goal,
+            end_time_goal,
+            duration_goal,
+            allow_running_forever,
+        )
 
         self._max_tasks = ntasks
 
-        self._pending_tasks = {}  # mapping from concurrent.futures.Future → point id
+        self._pending_tasks: dict[concurrent.Future, int] = {}
 
         # if we instantiate our own executor, then we are also responsible
         # for calling 'shutdown'
         self.shutdown_executor = shutdown_executor or (executor is None)
 
         self.learner = learner
-        self.log = [] if log else None
+        self.log: list | None = [] if log else None
 
         # Timing
         self.start_time = time.time()
-        self.end_time = None
+        self.end_time: float | None = None
         self._elapsed_function_time = 0
 
         # Error handling attributes
         self.retries = retries
         self.raise_if_retries_exceeded = raise_if_retries_exceeded
-        self._to_retry = {}
-        self._tracebacks = {}
+        self._to_retry: dict[int, int] = {}
+        self._tracebacks: dict[int, str] = {}
 
-        self._id_to_point = {}
-        self._next_id = functools.partial(
+        self._id_to_point: dict[int, Any] = {}
+        self._next_id: Callable[[], int] = functools.partial(
             next, itertools.count()
         )  # some unique id to be associated with each point
 
-    def _get_max_tasks(self):
+    def _get_max_tasks(self) -> int:
         return self._max_tasks or _get_ncores(self.executor)
 
-    def _do_raise(self, e, i):
-        tb = self._tracebacks[i]
-        x = self._id_to_point[i]
+    def _do_raise(self, e: Exception, pid: int) -> None:
+        tb = self._tracebacks[pid]
+        x = self._id_to_point[pid]
         raise RuntimeError(
             "An error occured while evaluating "
             f'"learner.function({x})". '
@@ -168,10 +242,10 @@ class BaseRunner(metaclass=abc.ABCMeta):
         ) from e
 
     @property
-    def do_log(self):
+    def do_log(self) -> bool:
         return self.log is not None
 
-    def _ask(self, n):
+    def _ask(self, n: int) -> tuple[list[int], list[float]]:
         pending_ids = self._pending_tasks.values()
         # using generator here because we only need until `n`
         pids_gen = (pid for pid in self._to_retry.keys() if pid not in pending_ids)
@@ -188,7 +262,7 @@ class BaseRunner(metaclass=abc.ABCMeta):
                 pids.append(pid)
         return pids, loss_improvements
 
-    def overhead(self):
+    def overhead(self) -> float:
         """Overhead of using Adaptive and the executor in percent.
 
         This is measured as ``100 * (1 - t_function / t_elapsed)``.
@@ -211,7 +285,10 @@ class BaseRunner(metaclass=abc.ABCMeta):
         t_total = self.elapsed_time()
         return (1 - t_function / t_total) * 100
 
-    def _process_futures(self, done_futs):
+    def _process_futures(
+        self,
+        done_futs: set[FutureTypes],
+    ) -> None:
         for fut in done_futs:
             pid = self._pending_tasks.pop(fut)
             try:
@@ -233,7 +310,9 @@ class BaseRunner(metaclass=abc.ABCMeta):
                     self.log.append(("tell", x, y))
                 self.learner.tell(x, y)
 
-    def _get_futures(self):
+    def _get_futures(
+        self,
+    ) -> list[FutureTypes]:
         # Launch tasks to replace the ones that completed
         # on the last iteration, making sure to fill workers
         # that have started since the last iteration.
@@ -255,7 +334,7 @@ class BaseRunner(metaclass=abc.ABCMeta):
         futures = list(self._pending_tasks.keys())
         return futures
 
-    def _remove_unfinished(self):
+    def _remove_unfinished(self) -> list[FutureTypes]:
         # remove points with 'None' values from the learner
         self.learner.remove_unfinished()
         # cancel any outstanding tasks
@@ -264,7 +343,7 @@ class BaseRunner(metaclass=abc.ABCMeta):
             fut.cancel()
         return remaining
 
-    def _cleanup(self):
+    def _cleanup(self) -> None:
         if self.shutdown_executor:
             # XXX: temporary set wait=True because of a bug with Python ≥3.7
             # and loky in any Python version.
@@ -276,8 +355,8 @@ class BaseRunner(metaclass=abc.ABCMeta):
         self.end_time = time.time()
 
     @property
-    def failed(self):
-        """Set of points that failed ``runner.retries`` times."""
+    def failed(self) -> set[int]:
+        """Set of points ids that failed ``runner.retries`` times."""
         return set(self._tracebacks) - set(self._to_retry)
 
     @abc.abstractmethod
@@ -293,15 +372,15 @@ class BaseRunner(metaclass=abc.ABCMeta):
         """Is called in `_get_futures`."""
 
     @property
-    def tracebacks(self):
+    def tracebacks(self) -> list[tuple[int, str]]:
         return [(self._id_to_point[pid], tb) for pid, tb in self._tracebacks.items()]
 
     @property
-    def to_retry(self):
+    def to_retry(self) -> list[tuple[int, int]]:
         return [(self._id_to_point[pid], n) for pid, n in self._to_retry.items()]
 
     @property
-    def pending_points(self):
+    def pending_points(self) -> list[tuple[FutureTypes, Any]]:
         return [
             (fut, self._id_to_point[pid]) for fut, pid in self._pending_tasks.items()
         ]
@@ -313,10 +392,26 @@ class BlockingRunner(BaseRunner):
     Parameters
     ----------
     learner : `~adaptive.BaseLearner` instance
-    goal : callable
+    goal : callable, optional
         The end condition for the calculation. This function must take
         the learner as its sole argument, and return True when we should
         stop requesting more points.
+    loss_goal : float, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the loss is smaller than this value.
+    npoints_goal : int, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the number of points is larger or
+        equal than this value.
+    end_time_goal : datetime, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than this
+        value.
+    duration_goal : timedelta or number, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than
+        ``start_time + duration_goal``. ``duration_goal`` can be a number
+        indicating the number of seconds.
     executor : `concurrent.futures.Executor`, `distributed.Client`,\
                `mpi4py.futures.MPIPoolExecutor`, `ipyparallel.Client` or\
                `loky.get_reusable_executor`, optional
@@ -369,34 +464,43 @@ class BlockingRunner(BaseRunner):
 
     def __init__(
         self,
-        learner,
-        goal,
+        learner: BaseLearner,
+        goal: Callable[[BaseLearner], bool] | None = None,
         *,
-        executor=None,
-        ntasks=None,
-        log=False,
-        shutdown_executor=False,
-        retries=0,
-        raise_if_retries_exceeded=True,
-    ):
+        loss_goal: float | None = None,
+        npoints_goal: int | None = None,
+        end_time_goal: datetime | None = None,
+        duration_goal: timedelta | int | float | None = None,
+        executor: (ExecutorTypes | None) = None,
+        ntasks: int | None = None,
+        log: bool = False,
+        shutdown_executor: bool = False,
+        retries: int = 0,
+        raise_if_retries_exceeded: bool = True,
+    ) -> None:
         if inspect.iscoroutinefunction(learner.function):
             raise ValueError("Coroutine functions can only be used with 'AsyncRunner'.")
         super().__init__(
             learner,
-            goal,
+            goal=goal,
+            loss_goal=loss_goal,
+            npoints_goal=npoints_goal,
+            end_time_goal=end_time_goal,
+            duration_goal=duration_goal,
             executor=executor,
             ntasks=ntasks,
             log=log,
             shutdown_executor=shutdown_executor,
             retries=retries,
             raise_if_retries_exceeded=raise_if_retries_exceeded,
+            allow_running_forever=False,
         )
         self._run()
 
-    def _submit(self, x):
+    def _submit(self, x: tuple[float, ...] | float | int) -> FutureTypes:
         return self.executor.submit(self.learner.function, x)
 
-    def _run(self):
+    def _run(self) -> None:
         first_completed = concurrent.FIRST_COMPLETED
 
         if self._get_max_tasks() < 1:
@@ -413,11 +517,11 @@ class BlockingRunner(BaseRunner):
                 concurrent.wait(remaining)
                 # Some futures get their result set, despite being cancelled.
                 # see https://github.com/python-adaptive/adaptive/issues/319
-                with_result = [f for f in remaining if not f.cancelled() and f.done()]
+                with_result = {f for f in remaining if not f.cancelled() and f.done()}
                 self._process_futures(with_result)
             self._cleanup()
 
-    def elapsed_time(self):
+    def elapsed_time(self) -> float:
         """Return the total time elapsed since the runner
         was started."""
         if self.end_time is None:
@@ -436,8 +540,25 @@ class AsyncRunner(BaseRunner):
     goal : callable, optional
         The end condition for the calculation. This function must take
         the learner as its sole argument, and return True when we should
-        stop requesting more points. If not provided, the runner will run
-        forever, or until ``self.task.cancel()`` is called.
+        stop requesting more points.
+        If not provided, the runner will run forever (or stop when no more
+        points can be added), or until ``runner.task.cancel()`` is called.
+    loss_goal : float, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the loss is smaller than this value.
+    npoints_goal : int, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the number of points is larger or
+        equal than this value.
+    end_time_goal : datetime, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than this
+        value.
+    duration_goal : timedelta or number, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than
+        ``start_time + duration_goal``. ``duration_goal`` can be a number
+        indicating the number of seconds.
     executor : `concurrent.futures.Executor`, `distributed.Client`,\
                `mpi4py.futures.MPIPoolExecutor`, `ipyparallel.Client` or\
                `loky.get_reusable_executor`, optional
@@ -462,6 +583,8 @@ class AsyncRunner(BaseRunner):
         the point is present in ``runner.failed``.
     raise_if_retries_exceeded : bool, default: True
         Raise the error after a point ``x`` failed `retries`.
+    allow_running_forever : bool, default: True
+        If True, the runner will run forever if the goal is not provided.
 
     Attributes
     ----------
@@ -500,22 +623,21 @@ class AsyncRunner(BaseRunner):
 
     def __init__(
         self,
-        learner,
-        goal=None,
+        learner: BaseLearner,
+        goal: Callable[[BaseLearner], bool] | None = None,
         *,
-        executor=None,
-        ntasks=None,
-        log=False,
-        shutdown_executor=False,
+        loss_goal: float | None = None,
+        npoints_goal: int | None = None,
+        end_time_goal: datetime | None = None,
+        duration_goal: timedelta | int | float | None = None,
+        executor: (ExecutorTypes | None) = None,
+        ntasks: int | None = None,
+        log: bool = False,
+        shutdown_executor: bool = False,
         ioloop=None,
-        retries=0,
-        raise_if_retries_exceeded=True,
-    ):
-
-        if goal is None:
-
-            def goal(_):
-                return False
+        retries: int = 0,
+        raise_if_retries_exceeded: bool = True,
+    ) -> None:
 
         if (
             executor is None
@@ -535,13 +657,18 @@ class AsyncRunner(BaseRunner):
 
         super().__init__(
             learner,
-            goal,
+            goal=goal,
+            loss_goal=loss_goal,
+            npoints_goal=npoints_goal,
+            end_time_goal=end_time_goal,
+            duration_goal=duration_goal,
             executor=executor,
             ntasks=ntasks,
             log=log,
             shutdown_executor=shutdown_executor,
             retries=retries,
             raise_if_retries_exceeded=raise_if_retries_exceeded,
+            allow_running_forever=True,
         )
         self.ioloop = ioloop or asyncio.get_event_loop()
         self.task = None
@@ -558,7 +685,7 @@ class AsyncRunner(BaseRunner):
             self.executor.shutdown()  # Make sure we don't shoot ourselves later
 
         self.task = self.ioloop.create_task(self._run())
-        self.saving_task = None
+        self.saving_task: asyncio.Task | None = None
         if in_ipynb() and not self.ioloop.is_running():
             warnings.warn(
                 "The runner has been scheduled, but the asyncio "
@@ -567,14 +694,14 @@ class AsyncRunner(BaseRunner):
                 "'adaptive.notebook_extension()'"
             )
 
-    def _submit(self, x):
+    def _submit(self, x: Any) -> asyncio.Task | asyncio.Future:
         ioloop = self.ioloop
         if inspect.iscoroutinefunction(self.learner.function):
             return ioloop.create_task(self.learner.function(x))
         else:
             return ioloop.run_in_executor(self.executor, self.learner.function, x)
 
-    def status(self):
+    def status(self) -> str:
         """Return the runner status as a string.
 
         The possible statuses are: running, cancelled, failed, and finished.
@@ -590,14 +717,21 @@ class AsyncRunner(BaseRunner):
         else:
             return "finished"
 
-    def cancel(self):
+    def cancel(self) -> None:
         """Cancel the runner.
 
         This is equivalent to calling ``runner.task.cancel()``.
         """
         self.task.cancel()
 
-    def live_plot(self, *, plotter=None, update_interval=2, name=None, normalize=True):
+    def live_plot(
+        self,
+        *,
+        plotter: Callable[[BaseLearner], holoviews.Element] | None = None,
+        update_interval: float = 2.0,
+        name: str = None,
+        normalize: bool = True,
+    ) -> holoviews.DynamicMap:
         """Live plotting of the learner's data.
 
         Parameters
@@ -621,10 +755,14 @@ class AsyncRunner(BaseRunner):
             The plot that automatically updates every `update_interval`.
         """
         return live_plot(
-            self, plotter=plotter, update_interval=update_interval, name=name
+            self,
+            plotter=plotter,
+            update_interval=update_interval,
+            name=name,
+            normalize=normalize,
         )
 
-    def live_info(self, *, update_interval=0.1):
+    def live_info(self, *, update_interval: float = 0.1) -> None:
         """Display live information about the runner.
 
         Returns an interactive ipywidget that can be
@@ -632,7 +770,7 @@ class AsyncRunner(BaseRunner):
         """
         return live_info(self, update_interval=update_interval)
 
-    async def _run(self):
+    async def _run(self) -> None:
         first_completed = asyncio.FIRST_COMPLETED
 
         if self._get_max_tasks() < 1:
@@ -650,7 +788,7 @@ class AsyncRunner(BaseRunner):
                 await asyncio.wait(remaining)
             self._cleanup()
 
-    def elapsed_time(self):
+    def elapsed_time(self) -> float:
         """Return the total time elapsed since the runner
         was started."""
         if self.task.done():
@@ -663,15 +801,26 @@ class AsyncRunner(BaseRunner):
             end_time = time.time()
         return end_time - self.start_time
 
-    def start_periodic_saving(self, save_kwargs, interval):
+    def start_periodic_saving(
+        self,
+        save_kwargs: dict[str, Any] | None = None,
+        interval: int = 30,
+        method: Callable[[BaseLearner], None] | None = None,
+    ):
         """Periodically save the learner's data.
 
         Parameters
         ----------
         save_kwargs : dict
             Key-word arguments for ``learner.save(**save_kwargs)``.
+            Only used if ``method=None``.
         interval : int
             Number of seconds between saving the learner.
+        method : callable
+            The method to use for saving the learner. If None, the default
+            saves the learner using "pickle" which calls
+            ``learner.save(**save_kwargs)``. Otherwise provide a callable
+            that takes the learner and saves the learner.
 
         Example
         -------
@@ -681,11 +830,19 @@ class AsyncRunner(BaseRunner):
         ...     interval=600)
         """
 
-        async def _saver(save_kwargs=save_kwargs, interval=interval):
+        def default_save(learner):
+            learner.save(**save_kwargs)
+
+        if method is None:
+            method = default_save
+            if save_kwargs is None:
+                raise ValueError("Must provide `save_kwargs` if method=None.")
+
+        async def _saver():
             while self.status() == "running":
-                self.learner.save(**save_kwargs)
+                method(self.learner)
                 await asyncio.sleep(interval)
-            self.learner.save(**save_kwargs)  # one last time
+            method(self.learner)  # one last time
 
         self.saving_task = self.ioloop.create_task(_saver())
         return self.saving_task
@@ -695,7 +852,15 @@ class AsyncRunner(BaseRunner):
 Runner = AsyncRunner
 
 
-def simple(learner, goal):
+def simple(
+    learner: BaseLearner,
+    goal: Callable[[BaseLearner], bool] | None = None,
+    *,
+    loss_goal: float | None = None,
+    npoints_goal: int | None = None,
+    end_time_goal: datetime | None = None,
+    duration_goal: timedelta | int | float | None = None,
+):
     """Run the learner until the goal is reached.
 
     Requests a single point from the learner, evaluates
@@ -710,10 +875,36 @@ def simple(learner, goal):
     Parameters
     ----------
     learner : ~`adaptive.BaseLearner` instance
-    goal : callable
-        The end condition for the calculation. This function must take the
-        learner as its sole argument, and return True if we should stop.
+    goal : callable, optional
+        The end condition for the calculation. This function must take
+        the learner as its sole argument, and return True when we should
+        stop requesting more points.
+    loss_goal : float, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the loss is smaller than this value.
+    npoints_goal : int, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the number of points is larger or
+        equal than this value.
+    end_time_goal : datetime, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than this
+        value.
+    duration_goal : timedelta or number, optional
+        Convenience argument, use instead of ``goal``. The end condition for the
+        calculation. Stop when the current time is larger or equal than
+        ``start_time + duration_goal``. ``duration_goal`` can be a number
+        indicating the number of seconds.
     """
+    goal = _goal(
+        learner,
+        goal,
+        loss_goal,
+        npoints_goal,
+        end_time_goal,
+        duration_goal,
+        allow_running_forever=False,
+    )
     while not goal(learner):
         xs, _ = learner.ask(1)
         for x in xs:
@@ -721,7 +912,10 @@ def simple(learner, goal):
             learner.tell(x, y)
 
 
-def replay_log(learner, log):
+def replay_log(
+    learner: BaseLearner,
+    log: list[tuple[Literal["tell"], Any, Any] | tuple[Literal["ask"], int]],
+) -> None:
     """Apply a sequence of method calls to a learner.
 
     This is useful for debugging runners.
@@ -737,10 +931,56 @@ def replay_log(learner, log):
         getattr(learner, method)(*args)
 
 
+# -- Internal executor-related, things
+
+
+def _ensure_executor(
+    executor: ExecutorTypes | None,
+) -> concurrent.Executor:
+    if executor is None:
+        executor = concurrent.ProcessPoolExecutor()
+
+    if isinstance(executor, concurrent.Executor):
+        return executor
+    elif with_ipyparallel and isinstance(executor, ipyparallel.Client):
+        return executor.executor()
+    elif with_distributed and isinstance(executor, distributed.Client):
+        return executor.get_executor()
+    else:
+        raise TypeError(
+            # TODO: check if this is correct. Isn't MPI,loky supported?
+            "Only a concurrent.futures.Executor, distributed.Client,"
+            " or ipyparallel.Client can be used."
+        )
+
+
+def _get_ncores(
+    ex: (ExecutorTypes),
+) -> int:
+    """Return the maximum  number of cores that an executor can use."""
+    if with_ipyparallel and isinstance(ex, ipyparallel.client.view.ViewExecutor):
+        return len(ex.view)
+    elif isinstance(
+        ex, (concurrent.ProcessPoolExecutor, concurrent.ThreadPoolExecutor)
+    ):
+        return ex._max_workers  # not public API!
+    elif isinstance(ex, loky.reusable_executor._ReusablePoolExecutor):
+        return ex._max_workers  # not public API!
+    elif isinstance(ex, SequentialExecutor):
+        return 1
+    elif with_distributed and isinstance(ex, distributed.cfexecutor.ClientExecutor):
+        return sum(n for n in ex._client.ncores().values())
+    elif with_mpi4py and isinstance(ex, mpi4py.futures.MPIPoolExecutor):
+        ex.bootup()  # wait until all workers are up and running
+        return ex._pool.size  # not public API!
+    else:
+        raise TypeError(f"Cannot get number of cores for {ex.__class__}")
+
+
 # --- Useful runner goals
 
-
-def stop_after(*, seconds=0, minutes=0, hours=0):
+# TODO: deprecate
+def stop_after(*, seconds=0, minutes=0, hours=0) -> Callable[[BaseLearner], bool]:
     """Stop a runner after a specified time.
 
     For example, to specify a runner that should stop after
@@ -774,63 +1014,126 @@ def stop_after(*, seconds=0, minutes=0, hours=0):
     return lambda _: time.time() > stop_time
 
 
-# -- Internal executor-related, things
+class _TimeGoal:
+    def __init__(self, dt: timedelta | datetime | int | float):
+        self.dt = dt if isinstance(dt, (timedelta, datetime)) else timedelta(seconds=dt)
+        self.start_time = None
+
+    def __call__(self, _):
+        if isinstance(self.dt, timedelta):
+            if self.start_time is None:
+                self.start_time = datetime.now()
+            return datetime.now() - self.start_time > self.dt
+        if isinstance(self.dt, datetime):
+            return datetime.now() > self.dt
+        raise TypeError(f"`dt={self.dt}` is not a datetime, timedelta, or number.")
 
 
-class SequentialExecutor(concurrent.Executor):
-    """A trivial executor that runs functions synchronously.
+def auto_goal(
+    *,
+    loss: float | None = None,
+    npoints: int | None = None,
+    end_time: datetime | None = None,
+    duration: timedelta | int | float | None = None,
+    learner: BaseLearner | None = None,
+    allow_running_forever: bool = True,
+) -> Callable[[BaseLearner], bool]:
+    """Extract a goal from the learners.
 
-    This executor is mainly for testing.
+    Parameters
+    ----------
+    loss : float, optional
+        Stop when the loss is smaller than this value.
+    npoints : int, optional
+        Stop when the number of points is larger or equal than this value.
+    end_time : datetime, optional
+        Stop when the current time is larger or equal than this value.
+    duration : timedelta or number, optional
+        Stop when the current time is larger or equal than
+        ``start_time + duration``. ``duration`` can be a number
+        indicating the number of seconds.
+    learner
+        Learner for which to determine the goal. Only used if the learner type
+        is `BalancingLearner`, `DataSaver`, `SequenceLearner`, or `IntegratorLearner`.
+    allow_running_forever
+        If True, and the goal is None and the learner is not a SequenceLearner,
+        then a goal that never stops is returned, otherwise an exception is raised.
+
+    Returns
+    -------
+    Callable[[adaptive.BaseLearner], bool]
     """
-
-    def submit(self, fn, *args, **kwargs):
-        fut = concurrent.Future()
-        try:
-            fut.set_result(fn(*args, **kwargs))
-        except Exception as e:
-            fut.set_exception(e)
-        return fut
-
-    def map(self, fn, *iterable, timeout=None, chunksize=1):
-        return map(fn, iterable)
-
-    def shutdown(self, wait=True):
-        pass
-
-
-def _ensure_executor(executor):
-    if executor is None:
-        executor = _default_executor()
-
-    if isinstance(executor, concurrent.Executor):
-        return executor
-    elif with_ipyparallel and isinstance(executor, ipyparallel.Client):
-        return executor.executor()
-    elif with_distributed and isinstance(executor, distributed.Client):
-        return executor.get_executor()
-    else:
-        raise TypeError(
-            "Only a concurrent.futures.Executor, distributed.Client,"
-            " or ipyparallel.Client can be used."
+    kw = dict(
+        loss=loss,
+        npoints=npoints,
+        end_time=end_time,
+        duration=duration,
+        allow_running_forever=allow_running_forever,
+    )
+    opts = (loss, npoints, end_time, duration)  # all are mutually exclusive
+    if sum(v is not None for v in opts) > 1:
+        raise ValueError(
+            "Only one of loss, npoints, end_time, duration can be specified."
         )
 
+    if loss is not None:
+        return lambda learner: learner.loss() <= loss
+    if isinstance(learner, BalancingLearner):
+        # Note that the float loss goal is more efficiently implemented in the
+        # BalancingLearner itself. That is why the previous if statement is
+        # above this one.
+        goals = [auto_goal(learner=l, **kw) for l in learner.learners]
+        return lambda learner: all(goal(l) for l, goal in zip(learner.learners, goals))
+    if npoints is not None:
+        return lambda learner: learner.npoints >= npoints
+    if end_time is not None:
+        return _TimeGoal(end_time)
+    if duration is not None:
+        return _TimeGoal(duration)
+    if isinstance(learner, DataSaver):
+        return auto_goal(**kw, learner=learner.learner)
+    if all(v is None for v in opts):
+        if isinstance(learner, SequenceLearner):
+            return SequenceLearner.done
+        if isinstance(learner, IntegratorLearner):
+            return IntegratorLearner.done
+        if not allow_running_forever:
+            raise ValueError(
+                "Goal is None which means the learners"
+                " continue forever and this is not allowed."
+            )
+        warnings.warn("Goal is None which means the learners continue forever!")
+        return lambda _: False
+    raise ValueError("Cannot determine goal from {goal}.")
 
-def _get_ncores(ex):
-    """Return the maximum  number of cores that an executor can use."""
-    if with_ipyparallel and isinstance(ex, ipyparallel.client.view.ViewExecutor):
-        return len(ex.view)
-    elif isinstance(
-        ex, (concurrent.ProcessPoolExecutor, concurrent.ThreadPoolExecutor)
+
+def _goal(
+    learner: BaseLearner | None,
+    goal: Callable[[BaseLearner], bool] | None,
+    loss_goal: float | None,
+    npoints_goal: int | None,
+    end_time_goal: datetime | None,
+    duration_goal: timedelta | None,
+    allow_running_forever: bool,
+):
+    if callable(goal):
+        return goal
+
+    if goal is not None and (
+        loss_goal is not None
+        or npoints_goal is not None
+        or end_time_goal is not None
+        or duration_goal is not None
     ):
-        return ex._max_workers  # not public API!
-    elif isinstance(ex, loky.reusable_executor._ReusablePoolExecutor):
-        return ex._max_workers  # not public API!
-    elif isinstance(ex, SequentialExecutor):
-        return 1
-    elif with_distributed and isinstance(ex, distributed.cfexecutor.ClientExecutor):
-        return sum(n for n in ex._client.ncores().values())
-    elif with_mpi4py and isinstance(ex, mpi4py.futures.MPIPoolExecutor):
-        ex.bootup()  # wait until all workers are up and running
-        return ex._pool.size  # not public API!
-    else:
-        raise TypeError(f"Cannot get number of cores for {ex.__class__}")
+        raise ValueError(
+            "Either goal, loss_goal, npoints_goal, end_time_goal or"
+            " duration_goal can be specified, not multiple."
+        )
+    return auto_goal(
+        learner=learner,
+        loss=loss_goal,
+        npoints=npoints_goal,
+        end_time=end_time_goal,
+        duration=duration_goal,
+        allow_running_forever=allow_running_forever,
+    )
